;; spager.jl -- Code to start spager
;; Copyright (C) 2000 Satyaki Das <satyaki@theforce.stanford.edu>
;;                    Ryan Lovett <ryan@ocf.Berkeley.EDU>
;;                    Andreas Buesching <crunchy@tzi.de>
;;                    Hakon Alstadheim <hakon.alstadheim@oslo.mail.telia.com>
;; $Id: spager.jl,v 1.37 2001/04/29 21:12:51 satyaki Exp $

;; This file is part of spager - a pager which works with sawfish.

;; spager is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; spager is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with spager; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;; Timers are needed for tooltips for windows
(require 'timers)
;; new x-message package from merlin
(require 'merlin.message)
;; need become of new internal structure
(require 'sawfish.wm.workspace)
(require 'sawfish.wm.util.window-order)

;; Source code directory
(defvar spager-directory "/sprout/sprout3/satyaki/spager")

;;; Customization code contributed by Ryan Lovett <ryan@ocf.Berkeley.EDU>
(defgroup spager "Spager")

(defcustom spager-show-current-workspace-only nil
  "Show only current workspace"
  :type boolean
  :group spager
  :after-set (lambda ()
	       (setq spager-num-workspaces -1)
	       (spager-send-resize-data)))

(defcustom spager-workspaces-per-column 1
  "The number of workspaces per column."
  :type number
  :group spager
  :after-set (lambda () (spager-set-workspaces-per-column
			 spager-workspaces-per-column)))

(defcustom spager-viewport-width 40
  "The width of each mini viewport/workspace"
  :type number
  :group spager
  :after-set (lambda ()
	       (setq spager-num-workspaces -1)
	       (spager-send-resize-data)))

(defcustom spager-viewport-height 30
  "The height of each mini viewport/workspace"
  :type number
  :group spager
  :after-set (lambda ()
	       (setq spager-viewport-height spager-viewport-height)
	       (setq spager-num-workspaces -1)
	       (spager-send-resize-data)))

(defcustom spager-color-background (get-color "#dddddd")
  "Pager background"
  :type color
  :group spager
  :after-set (lambda ()
	       (spager-change-color 'background spager-color-background)
	       (spager-set-colors)))

(defcustom spager-color-hilit-background (get-color "#999999")
  "Current viewport/workspace background"
  :type color
  :group spager
  :after-set (lambda ()
	       (spager-change-color 'hilit-background
				    spager-color-hilit-background)
	       (spager-set-colors)))

(defcustom spager-color-window (get-color "#006699")
  "Windows"
  :type color
  :group spager
  :after-set (lambda ()
	       (spager-change-color 'window spager-color-window)
	       (spager-set-colors)))

(defcustom spager-color-focus-window (get-color "#55aaff")
  "Window with input focus"
  :type color
  :group spager
  :after-set (lambda ()
	       (spager-change-color 'focus-window spager-color-focus-window)
	       (spager-set-colors)))

(defcustom spager-color-viewport-divider (get-color "#544454445444")
  "Lines separating viewports"
  :type color
  :group spager
  :after-set (lambda ()
	       (spager-change-color 'viewport-divider
				    spager-color-viewport-divider)
	       (spager-set-colors)))

(defcustom spager-color-workspace-divider (get-color "#000000")
  "Lines separating workspaces"
  :type color
  :group spager
  :after-set (lambda ()
	       (spager-change-color 'workspace-divider
				    spager-color-workspace-divider)
	       (spager-set-colors)))

(defcustom spager-enable-window-labels t
  "Enable display of window labels"
  :type boolean
  :group spager)

(defcustom spager-delay 900
  "Delay in ms before window label shows up"
  :type number
  :group spager)

(defcustom spager-label-font (get-font "fixed")
  "Font to use for the window label"
  :type font
  :group spager)

(defcustom spager-label-background (get-color "white")
  "Window label background"
  :type color
  :group spager)

(defcustom spager-label-x-offset -30
  "Horizontal offset of label from mouse position"
  :type number
  :group spager
  :range (-500 . 500))

(defcustom spager-label-y-offset 30
  "Vertical offset of label from mouse position"
  :type number
  :group spager
  :range (-500 . 500))

;; Tooltip timer...
(defvar spager-timer nil)

;; Size of the pager windows...
(defvar spager-executable (format nil "%s/spager" spager-directory))
(defvar spager-colors
  `(;;Pager background
    (background ,@(color-rgb spager-color-background))
    ;;Current viewport background
    (hilit-background ,@(color-rgb spager-color-hilit-background))
    ;;Windows
    (window ,@(color-rgb spager-color-window))
    ;;Window with input focus
    (focus-window ,@(color-rgb spager-color-focus-window))
    ;;Lines separating the viewports
    (viewport-divider ,@(color-rgb spager-color-viewport-divider))
    ;;Lines separating the workspaces
    (workspace-divider ,@(color-rgb spager-color-workspace-divider)))
  "Describes the colors used in the pager. No error checking so should be
in the order above.")

;; Remembers the number of workspaces...
(defvar spager-num-workspaces (let ((val (workspace-limits)))
				  (+ 1 (- (cdr val) (car val)))))

;; Remembers the viewport dimensions...
(defvar spager-viewport-rows (cdr viewport-dimensions))
(defvar spager-viewport-columns (car viewport-dimensions))

(defvar spager-width nil)
(defvar spager-height nil)
(defvar spager-process nil)
(defvar spager-hooks
  '((viewport-moved-hook . spager-send-viewport-state)
    (after-resize-hook . spager-send-window-state)
    (after-move-hook . spager-send-window-state)
    (window-maximized-hook . spager-send-window-state)
    (window-unmaximized-hook . spager-send-window-state)
    (map-notify-hook . spager-send-window-state)
    (unmap-notify-hook . spager-send-window-state)
    (unmap-notify-hook . spager-delete-sticky-window)
    (window-state-change-hook . spager-manage-sticky-windows)
    (iconify-window-hook . spager-send-window-state)
    (uniconify-window-hook . spager-send-window-state)
    (shade-window-hook . spager-send-window-state)
    (unshade-window-hook . spager-send-window-state)
    (after-restacking-hook . spager-send-state)
    (focus-in-hook . spager-send-focus-state)
    (focus-out-hook . spager-send-focus-state)
    (enter-workspace-hook . spager-send-viewport-state)
    (workspace-state-change-hook . spager-send-resize-data)
    (viewport-resized-hook . spager-send-resize-data))
  "These are the hooks to which we add the different display functions.")

(defvar spager-sticky-windows ()
  "Maintain a list of sticky windows ids.")

;; For old versions of sawmill
(when (not (boundp 'floor))
  (setq floor (lambda (x) x)))

(defun spager-change-color (keyword color)
  "Changes a specific color"
  (setq spager-colors
	(mapcar #'(lambda (x)
		    (if (eq (car x) keyword)
			`(,keyword ,@(color-rgb color))
			x))
		spager-colors)))

(defun spager-compute-size ()
  "Does size calculations for the pager window."
  (setq spager-num-workspaces 
	(if spager-show-current-workspace-only
	    1
	    (+ 1 (cdr (workspace-limits)) (- (car (workspace-limits))))))
  (setq spager-viewport-rows (cdr viewport-dimensions))
  (setq spager-viewport-columns (car viewport-dimensions))
  ;; Todd Larason reminded me that ceiling is not available in older
  ;; versions of librep. Thanks Todd!
  (setq spager-width (* (car viewport-dimensions)
			spager-viewport-width
			(if (boundp 'ceiling)
			    (ceiling (/ spager-num-workspaces
					spager-workspaces-per-column))
			    (+ (/ spager-num-workspaces
				  spager-workspaces-per-column)
			       (if (> (% spager-num-workspaces
					 spager-workspaces-per-column)
				      0)
				   1 0)))))
  (setq spager-height (* (cdr viewport-dimensions)
			 spager-viewport-height
			 (if (>= spager-workspaces-per-column
				 spager-num-workspaces)
			     spager-num-workspaces
			     spager-workspaces-per-column))))

(defun spager-set-colors (#!optional c)
  "Tells the C program to change the colors to whatever is in the
`SPAGER-COLORS' variable."
  (when spager-process
    (format spager-process
	    "%s\n"
	    (concat
	     (if c "" "c")
	     (apply #'(lambda (first #!rest args)
			(apply concat
			       (substring first 1 (length first))
			       args))
		    (mapcar #'(lambda (desc)
				(format nil " %s %s %s"
					(cadr desc) (caddr desc)
					(car (cdddr desc))))
			    spager-colors))))))

(defun spager-set-workspaces-per-column (value)
  "Allows you to change the number of workspaces per column."
  (setq spager-workspaces-per-column value)
  (setq spager-num-workspaces -1) ;Force send-resize-data to communicate
				    ;with the C program.
  (spager-send-resize-data))

(defun spager-start ()
  "This is the function that you should call to start the pager."
  (spager-compute-size)
  ;; If the pager is already running kill it and remove the hooks that we
  ;; inserted.
  (when spager-process
    (kill-process spager-process))
  (spager-remove-hooks)
  (setq spager-sticky-windows
	(mapcar #'window-id
		(remove-if #'(lambda (w)
			       (and (not (window-get w 'sticky))
				    (not (window-get w 'viewport-sticky))))
			   (stacking-order))))
  (spager-start-process)
  (spager-send-init-data)
  (spager-set-colors t)
  (spager-add-hooks)
  (spager-send-state))

(defun spager-stop ()
  "This is the function that you should call to get rid of the pager."
  (when spager-process
    (kill-process spager-process))
  (spager-remove-hooks))

(defun spager-add-hooks ()
  "This function adds hooks to various events that should change the pager
display, so the pager display gets updated."
  (condition-case err-info
      (mapc (lambda (hook)
		(add-hook (car hook) (symbol-value (cdr hook)) t))
	    spager-hooks)
    (error
     (format standard-error "spager: error in add-hooks %s\n" err-info))))

(defun spager-remove-hooks ()
  "This function removes the hooks that we added to sawmill."
  (condition-case err-info
      (mapc (lambda (hook)
	      (and (boundp (car hook) )
		   (remove-hook (car hook) (symbol-value (cdr hook)))))
	    spager-hooks)
    (error
     (format standard-error "spager: error in remove-hooks %s\n" err-info))))

(defun spager-convert (x y)
  "Given a location in the pager window, this translates into a location
in the actual workspaces. Returns a list of 3 numbers. The first is the
workspace number while the remaining two are the location relative to the
start of the workspace."
  (let ((ws-x (floor (/ x (* (car viewport-dimensions)
			     spager-viewport-width))))
	(ws-y (floor (/ y (* (cdr viewport-dimensions)
			     spager-viewport-height)))))
    (list (+ ws-y (* spager-workspaces-per-column ws-x))
	  (spager-scale (mod x (* (car viewport-dimensions)
				  spager-viewport-width))
			(screen-width)
			spager-viewport-width)
	  (spager-scale (mod y (* (cdr viewport-dimensions)
				  spager-viewport-height))
			(screen-height)
			spager-viewport-height))))

(defun spager-goto (x y)
   "This is the function which the C program calls to change viewports or
workspaces. The co-ords are what the C program thinks are the co-ordinates
were the user clicked."
  (spager-remove-hooks)
  (let* ((loc1 (spager-convert x y))
	 (loc2 (list (car loc1)
		     (floor (/ (cadr loc1) (screen-width)))
		     (floor (/ (caddr loc1) (screen-height))))))
    (if (eql (car loc2) current-workspace)
	(apply set-screen-viewport (cdr loc2))
      (apply select-workspace-and-viewport loc2)))
  (spager-add-hooks)
  (spager-send-viewport-state))


(defun spager-start-process ()
  "Start a pager process. Inspired by John's code for the menu process. I
just skimmed the code, so I may have missed subtle points."
  (let ((sentinel (lambda ()
		    (when (and spager-process
			       (not (process-in-use-p spager-process)))
		      (setq spager-process nil))))
	(filter nil))
    (setq spager-process (make-process filter sentinel)))
  (start-process spager-process spager-executable))

(defun spager-send-init-data ()
  "Initially send the width and height of the pager window to be displayed
by the C process. For now I don't allow the size to be changed on the fly."
  (when spager-process
    (format spager-process "%d %d %d %d %d %d\n"
	    spager-viewport-width
	    spager-viewport-height
	    (* spager-viewport-width (car viewport-dimensions))
	    (* spager-viewport-height (cdr viewport-dimensions))
	    spager-width
	    spager-height)))

(defun spager-send-resize-data ()
  "This tells the pager what the new size of the window is going to be."
  (unless (and (eql spager-num-workspaces
		    (let ((val (workspace-limits)))
		      (+ 1 (- (cdr val) (car val)))))
	       (eql (cdr viewport-dimensions) spager-viewport-rows)
	       (eql (car viewport-dimensions) spager-viewport-columns))
    (spager-compute-size)
    (when spager-process
      (format spager-process "s%d %d %d %d %d %d\n"
	      spager-viewport-width
	      spager-viewport-height
	      (* spager-viewport-width (car viewport-dimensions))
	      (* spager-viewport-height (cdr viewport-dimensions))
	      spager-width
	      spager-height))
    (spager-send-state)))

(defun spager-scale (val factor1 factor2)
  "This is just to scale the window co-ords and dimensions."
  (if (boundp 'round)
      (round (/ (* val factor1) factor2))
      (let ((mult (* val factor1)))
	(+ (if (> (mod mult factor2) (/ factor2 2)) 1 0)
	   (/ mult factor2)))))

(defun spager-move-window-absolute (window x y)
  "Moves the window to the position co-ordinates (X, Y) relative to the
upper left hand corner of the workspace."
  (move-window-to window
		  (- x viewport-x-offset)
		  (- y viewport-y-offset)))

(defun spager-move-window (id x y)
  "Moves window with id ID to co-ordinates (X, Y) where (X, Y) is what the
pager thinks the position of the window is. This is the function that the
pager calls."
  ;; First do bounds checking...
  (let* ((x (if (< x 0) 0 x))
	 (x (if (>= x (- spager-width 5)) (- spager-width 5) x))
	 (y (if (< y 0) 0 y))
	 (y (if (>= y (- spager-height 5)) (- spager-height 5) y))
	 (location (spager-convert x y))
	 (win (get-window-by-id id))
	 (cws (and win (window-get win 'workspaces))))
    ;(when (equal sawmill-version "0.25.2")
    ;  (hide-window win))
    (spager-move-window-absolute 
     win
     (spager-scale (mod x (* (car viewport-dimensions) spager-viewport-width))
		   (screen-width) spager-viewport-width)
     (spager-scale (mod y (* (cdr viewport-dimensions) spager-viewport-height))
		   (screen-width) spager-viewport-width))
    ;; Don't change workspace if the window is sticky or present in
    ;; multiple workspaces or we are in single workspace mode
    (when (and win cws (null (cdr cws)) (not (eql (car location) (car cws)))
	       (not spager-show-current-workspace-only))
      (spager-window-to-workspace win (car location) nil))
    ;(when (equal sawmill-version "0.25.2")
    ;  (show-window win))
    (spager-send-window-state)))

;; This should be fixed with newer versions of sawmill?!
(defun spager-window-to-workspace (w count #!optional copy)
  "Same as send-window-to-workspace-from-first except doesn't change
current workspace."
  (let* ((was-focused (eq (input-focus) w))
	 (orig-space (if (window-in-workspace-p w current-workspace)
			 current-workspace
			 (car (window-workspaces w))))
	 (new-space (workspace-id-from-logical count)))
    (when orig-space
      (ws-copy-window w orig-space new-space nil)
      (when (eql orig-space current-workspace)
	(delete-window-instance w))
      (spager-send-window-state)
      (unless copy
	(ws-move-window w orig-space new-space was-focused)))))

(defun spager-send-state (#!rest args)
  "Tell the C program what to display. Since I wrote a C function to
read the data, the format is very strict. The first pair of integers is
the viewport currently focussed. The third number in the sequence is the
active window index (and a larger number if there are none). Finally for
each window displayed we need to send four numbers (the co-ords of upper
left corner and the width and height of the window). The string sent
contains integers separated from each other by exactly one space and the
string is terminated by a newline. So for instance the following is a
valid string to send to the pager process:

    \"d0 0 67108866 35 57 2 2 37 57 2 2 32 57 2 2\n\"

Putting extra spaces anywhere in between will have *bad* consequences
because my C routine is fragile.

*** Warning: This documentation is really out of date:
As far as I can decipher the first three integers are okay. After that for
each window we send seven (not four) integers. The first four integers are
the position and dimensions. The fifth is the window id. The sixth and
seventh have something to do with displaying sticky windows in every viewport."
  ;(ws-normalize-indices)
  (when spager-process
    (format
     spager-process "%s\n"
     (apply concat
	    (format nil "d%s %s %s"
		    (+ (floor (/ viewport-x-offset (screen-width)))
		       (* (car viewport-dimensions)
			  (floor (/ (if spager-show-current-workspace-only
					0 
					current-workspace)
				    spager-workspaces-per-column))))
		    (+ (floor (/ viewport-y-offset (screen-height)))
		       (* (mod (if spager-show-current-workspace-only
				   0
				   current-workspace)
			       spager-workspaces-per-column)
			  (cdr viewport-dimensions)))
		    (if (input-focus)
			(window-id (input-focus))
			0))
	    (mapcar spager-get-window-info 
		    (spager-filter (reverse (stacking-order))))))))

(defun spager-send-viewport-state (#!rest args)
  "Send the viewport that is in focus. Then send the new position of
sticky windows only."
  (when spager-process
    (format spager-process
	    "v%s %s\n"
	    (+ (floor (/ viewport-x-offset (screen-width)))
	       (* (car viewport-dimensions)
		  (floor (/ (if spager-show-current-workspace-only
				0 
				current-workspace)
			    spager-workspaces-per-column))))
	    (+ (floor (/ viewport-y-offset (screen-height)))
	       (* (mod (if spager-show-current-workspace-only
			   0
			   current-workspace)
		       spager-workspaces-per-column)
		  (cdr viewport-dimensions))))
    (mapc #'(lambda (id) (spager-send-window-state (get-window-by-id id)))
	  spager-sticky-windows)))

(defun spager-send-window-state (w #!rest args)
  "When only the size or iconification or shading of a window changes send
pnly the data pertaining to that window."
  (when spager-process
    (format spager-process "w%s\n" (substring (spager-get-window-info w) 1))))

(defun spager-send-focus-state (#!rest args)
  "Sends window-id of focussed window (or 0 if none is focussed)."
  (when spager-process
    (format spager-process "f%s\n" (if (input-focus)
				       (window-id (input-focus))
				       0))))

(defun spager-filter (win-list)
  (if spager-show-current-workspace-only
      (apply append
	     (mapcar (lambda (w)
		       (let ((ws (window-get w 'workspaces)))
			 (if (or (null ws)
				 (member current-workspace ws))
			     (list w)
			     ())))
		     win-list))
      win-list))

(defun spager-get-window-info (w)
  "Allows for copies of the same window to be there in different
workspaces."
  (apply 
   concat
   (mapcar
    (lambda (wid)
      (let ((pos (spager-window-position w wid))
	    (dim-orig (window-dimensions w))
	    (dim (spager-window-dimensions w wid)))
	(if (and (not (window-get w 'iconified))
		 (window-mapped-p w))
	    (format nil " %d %d %d %d %d %d %d"
		    (spager-scale (car pos) spager-viewport-width
				  (screen-width))
		    (spager-scale (cdr pos) spager-viewport-height
				  (screen-height))
		    (spager-scale (car dim) spager-viewport-width
				  (screen-width))
		    (if (window-get w 'shaded)
			2
			(spager-scale (cdr dim) spager-viewport-height
				      (screen-height)))
		    (or (window-id w) 0)
		    (spager-scale (car dim-orig) spager-viewport-width
				  (screen-width))
		    (if (window-get w 'shaded)
			2
			(spager-scale (cdr dim-orig) spager-viewport-height
				      (screen-height))))
	    (format nil " %d %d 0 0 %d 0 0" spager-width spager-height
		    (or (window-id w) 0)))))
    (or (window-get w 'workspaces)
	(list current-workspace)))))

(defun spager-window-dimensions (window wid)
  (let* ((pos-corrected (spager-window-position window wid))
	 (pos (cons (- (car pos-corrected)
		       (* (screen-width) (car viewport-dimensions)
			  (floor (/ (- wid (car (workspace-limits)))
				    spager-workspaces-per-column))))
		    (- (cdr pos-corrected)
		       (* (screen-height) (cdr viewport-dimensions)
			  (mod (- wid (car (workspace-limits)))
			       spager-workspaces-per-column)))))
	 (dim (window-dimensions window)))
    (cons (if (> (+ (car pos) (car dim))
		 (* (screen-width) (car viewport-dimensions)))
	      (- (* (screen-width) (car viewport-dimensions)) (car pos))
	      (car dim))
	  (if (> (+ (cdr pos) (cdr dim))
		 (* (screen-height) (cdr viewport-dimensions)))
	      (- (* (screen-height) (cdr viewport-dimensions)) (cdr pos))
	      (cdr dim)))))

(defun spager-window-position (window wid)
  "This function is to return the position of the ``WINDOW'' in the
current workspace. For now sticky windows show up only in the current
window. This is consistent with the thing that fvwm2 does. This function
is very similar to ``WINDOW-POSITION'' except that the positions are given
with respect to the upper left corner of the workspace (and not the
current viewport).

We have also added an adjustment so that the position is corrected with
respect to where it should be because of multiple workspaces.

TODO: Have an option for sticky windows to appear in all windows."
  (let* ((pos (window-position window))
	 (vp (screen-viewport))
	 (ws-x (if spager-show-current-workspace-only
		   0
		   (floor (/ (- wid (car (workspace-limits)))
			     spager-workspaces-per-column))))
	 (ws-y (if spager-show-current-workspace-only
		   0
		   (mod (- wid (car (workspace-limits)))
			spager-workspaces-per-column))))
    (cons (+ (car pos) (* (car vp) (screen-width))
	     (* (screen-width) (car viewport-dimensions) ws-x))
	  (+ (cdr pos) (* (cdr vp) (screen-height))
	     (* (screen-height) (cdr viewport-dimensions) ws-y)))))

(defun spager-manage-sticky-windows (w)
  (let ((id (window-id w)))
    (cond ((or (window-get w 'sticky)
	       (window-get w 'viewport-sticky))
	   (unless (member id spager-sticky-windows)
	     (setq spager-sticky-windows (cons id spager-sticky-windows))))
	  (t (setq spager-sticky-windows (remove id spager-sticky-windows))))))

(defun spager-delete-sticky-window (w)
  (setq spager-sticky-windows (remove (window-id w) spager-sticky-windows)))

(defun spager-start-timer (id x y)
  "This is the function the C program calls. This should get rid of the
timer if it is already set and restart it."
  (when spager-enable-window-labels
    (when spager-timer
      (delete-timer spager-timer)
      (setq spager-timer nil))
    (hide-fancy-message)
    (setq spager-timer (make-timer (lambda () (spager-show-tooltip id x y))
				   0 (+ spager-delay 100)))))

(defun spager-stop-timer ()
  "Remove message and delete timer."
  (when spager-enable-window-labels
    (when spager-timer
      (delete-timer spager-timer)
      (setq spager-timer nil))
    (hide-fancy-message)))

(defun spager-show-tooltip (id x y)
  (let* ((pos (window-position (get-window-by-name "spager")))
	 (pos (cons (- (+ (car pos) x) spager-label-x-offset)
		    (- (+ (cdr pos) y) spager-label-y-offset))))
    (fancy-message
     `(,(window-name (get-window-by-id id)))
     `((position . ,pos)
       (font . ,spager-label-font)
       (background . ,spager-label-background)
       (gravity . none)))))

(defun spager-change-depth (id)
  (and (get-window-by-id id)
       (raise-lower-window (get-window-by-id id))))

(provide 'spager)
